
//lib for ESP8266
#include "ESP8266WiFi.h"

//lib for  MQTT
#include "Adafruit_MQTT.h"
#include "Adafruit_MQTT_Client.h"

//lib for DHT11
#include "DHT.h"

//lib for JSON message
#include "ArduinoJson.h"

/************************* WiFi Access Point *********************************/

#define WLAN_SSID       ""
#define WLAN_PASS       ""

/************************* MQTT deklarace *********************************/

#define AIO_SERVER      ""
#define AIO_SERVERPORT  1883
#define AIO_USERNAME    ""
#define AIO_KEY         ""

/************************* Mapování PINů *********************************/

const byte DHTPIN = 4;
const byte WSPEED = 14;
const byte WDIR = A0;

/************************* Globální proměnné *********************************/
char msg[100];

byte mac[6];
char znaky[1];
String adresaMAC;
String adresaIP;
String nazevZarizeni;
int silaSignal;
char MQTTRegistraceJSON[200];

char MQTTMeteoVitrJSON[200];
long lastSecond; //The millis windClicks to see when a second rolls by
long lastSecondSpanek; //The millis windClicks to see when a second rolls by
long lastSecondPrint; //The millis windClicks to see when a second rolls by TISK
byte seconds; //When it hits 60, increase the current minute
unsigned int seconds_5m; //Keeps track of the "wind speed/dir avg" over last 5 minutes array of data
byte minutes; //Keeps track of where we are in various arrays of data
byte minutes_5m; //Keeps track of where we are in wind gust/dir over last 5 minutes array of data

float speedwind_ms = 0;

long lastWindCheck = 0;
volatile byte windClicks = 0;

//We need to keep track of the following variables:
//Wind speed/dir each update (no storage)
//Wind gust/dir over the day (no storage)
//Wind speed/dir, avg over 2 minutes (store 1 per second)
//Wind gust/dir over last 5 minutes (store 1 per minute)

#define WIND_DIR_AVG_SIZE 300
float windspdavg[WIND_DIR_AVG_SIZE]; //300 bytes to keep track of 5 minute average
unsigned int winddiravg[WIND_DIR_AVG_SIZE]; //300 ints to keep track of 5 minute average

float windgust_5m[5]; //5 floats to keep track of 5 minute max
unsigned int windgustdirection_5m[5]; //5 ints to keep track of 5 minute max
volatile float rainHour[60]; //60 floating numbers to keep track of 60 minutes of rain

//These are all the weather values that wunderground expects:
unsigned int winddir = 0; // [0-360 instantaneous wind direction]
float windspeedms = 0; // [m/s instantaneous wind speed]
float windgustms = 0; // [m/s current wind gust, using software specific time period]
unsigned int windgustdir = 0; // [0-360 using software specific time period]
float windspdms_avg5m = 0; // [m/s 2 minute average wind speed m/s]
unsigned int winddir_avg5m = 0; // [0-360 2 minute average wind direction]
float windgustms_5m = 0; // [m/s past 5 minutes wind gust m/s ]
unsigned int windgustdir_5m = 0; // [0-360 past 5 minutes wind gust direction]
float humidity = 0; // [%]
float tempC = 0; // [temperature C°]

/************************* BHT11 deklarace *********************************/

#define DHTTYPE DHT11   // DHT 11

// Inicializace DHT
DHT dht(DHTPIN, DHTTYPE);

/****************************** Nastavení Wifi a MQTT ***************************************/

// Create an ESP8266 WiFiClient class to connect to the MQTT server.
WiFiClient client;

// Setup the MQTT client class by passing in the WiFi client and MQTT server and login details.
Adafruit_MQTT_Client mqtt(&client, AIO_SERVER, AIO_SERVERPORT, AIO_USERNAME, AIO_KEY);

//publish
Adafruit_MQTT_Publish meteo_teplota = Adafruit_MQTT_Publish(&mqtt, AIO_USERNAME "pozemek/anemometer/temp");
Adafruit_MQTT_Publish meteo_vlhkost = Adafruit_MQTT_Publish(&mqtt, AIO_USERNAME "pozemek/anemometer/hum");
Adafruit_MQTT_Publish meteo_vitr_smer = Adafruit_MQTT_Publish(&mqtt, AIO_USERNAME "pozemek/anemometer/winddir");
Adafruit_MQTT_Publish meteo_vitr_rychlost = Adafruit_MQTT_Publish(&mqtt, AIO_USERNAME "pozemek/anemometer/windspeed");
Adafruit_MQTT_Publish meteo_vitr_json = Adafruit_MQTT_Publish(&mqtt, AIO_USERNAME "pozemek/anemometer/wind_json");
Adafruit_MQTT_Publish sit_registrace_pridat = Adafruit_MQTT_Publish(&mqtt, AIO_USERNAME "sit/registrace/reg");

void WiFiOnRegistrace();
void WiFiOn();
void WiFiOff();
void MQTT_connect();

//windClicks
ICACHE_RAM_ATTR void wspeedIRQ()
{
  windClicks++; //There is 1.492MPH for each click per second.
}

void setup() {

  Serial.begin(9600);

  //Přerušení pro rychlost větru
  pinMode(WSPEED, INPUT_PULLUP); // input from wind meters windspeed sensor
  attachInterrupt(digitalPinToInterrupt(WSPEED), wspeedIRQ, FALLING);
  interrupts();

  // DHT11
  dht.begin();

//  Serial.println("MQTT ESP8266 - sit_registrace_pridat");
  WiFiOnRegistrace();
  MQTT_connect();
  
  //ping the server to keep the mqtt connection alive
  if(!mqtt.ping()) {
    mqtt.disconnect();
  }

  sit_registrace_pridat.publish(MQTTRegistraceJSON);
  
  WiFiOff();
  
  seconds = 0;
  seconds_5m = 0;
  minutes = 0;
  minutes_5m = 0;

  lastSecond = millis();
}

void loop() {  
  //Keep track of which minute it is
  if(millis() - lastSecond >= 1000) 
  {
        lastSecond += 1000;

        //Calc the wind speed and direction every second for 300 second to get 5 minute average
        float currentSpeed = get_wind_speed();
        //float currentSpeed = random(1290); //For testing
        
        windspeedms = currentSpeed;//update global variable for windspeed when using the publikovatHodnoty() function

        int currentDirection = get_wind_direction();
        //int currentDirection = random(0,360); //For testing
        
        windspdavg[seconds_5m] = currentSpeed;
        winddiravg[seconds_5m] = currentDirection;

        //Check to see if this is a gust for the minute
        if(currentSpeed > windgust_5m[minutes_5m])
        {
            windgust_5m[minutes_5m] = currentSpeed;
            windgustdirection_5m[minutes_5m] = currentDirection;
        }

        //Check to see if this is a gust for the day
        if(currentSpeed > windgustms)
        {
            windgustms = currentSpeed;
            windgustdir = currentDirection;
        }

        if(++seconds > 59)
        {
            seconds = 0;

            if(++minutes_5m > 4) {
              minutes_5m = 0;
              windgust_5m[minutes_5m] = 0; //Zero out this minute's gust

              publikovatHodnotyPredSpanim();
              delay(1200);
              
              //Serial.println("10 minutes sleep");
              //ESP.deepSleep(600e6); // 
              Serial.println("5 minutes sleep");
              ESP.deepSleep(300e6); // 
              //Serial.println("20 seconds sleep for testing");
              //ESP.deepSleep(20e6); // 20e6 is 20 microseconds 60e6 is 60 sekund
           }              
        }

        if(++seconds_5m > 299) seconds_5m = 0;
    }

  delay(100);
  
}

//Funkce pro zapínání a vypínání wifi modulu
void WiFiOnRegistrace() {
  WiFi.forceSleepWake();
  delay(1);
  
  WiFi.mode(WIFI_STA);
  
  Serial.println("ESP8266 - pripojeni WiFi a registracni JSON");

  // Connect to WiFi access point.
  Serial.print("Pripojuji se k ");
  Serial.println(WLAN_SSID);

  WiFi.begin(WLAN_SSID, WLAN_PASS);
  while (WiFi.status() != WL_CONNECTED) {
    delay(500);
    Serial.print(".");
  }
  Serial.println();

  Serial.println("WiFi Pripojeno!");
  Serial.println("IP adresa: "); Serial.println(WiFi.localIP());
  
  //Registration 
  //Zjistit MAC adresu
  WiFi.macAddress(mac);

  for (int i=0; i<6; i++) {
      sprintf(znaky, "%02x%s", mac[i], i != 5 ? ":" : "");
      adresaMAC += znaky;
  }

  //Zjistit IP adresu
  adresaIP = WiFi.localIP().toString();

  //Název zařízení
  nazevZarizeni = "ANEMOMETER_MQTT_Pub_VITR";
  WiFi.hostname(nazevZarizeni);

  //Síla signálu
  silaSignal = WiFi.RSSI();
  
  //Vytvoření JSON po odeslání informací o mě
  StaticJsonBuffer<200> jsonBuffer;
  JsonObject& registraceJSON = jsonBuffer.createObject();

  registraceJSON["MAC"] = adresaMAC;
  registraceJSON["nazev"] = nazevZarizeni;
  registraceJSON["IP"] = adresaIP;
  registraceJSON["signal"] = silaSignal;

  registraceJSON.printTo(MQTTRegistraceJSON);
  registraceJSON.printTo(Serial);
  Serial.println();
}

void WiFiOn() {
  WiFi.forceSleepWake();
  delay(1);
  
  WiFi.mode(WIFI_STA);
  
  Serial.println("ESP8266 - pripojeni WiFi a registracni JSON");

  // Connect to WiFi access point.
  Serial.print("Pripojuji se k ");
  Serial.println(WLAN_SSID);

  WiFi.begin(WLAN_SSID, WLAN_PASS);
  while (WiFi.status() != WL_CONNECTED) {
    delay(500);
    Serial.print(".");
  }
  Serial.println();

  Serial.println("WiFi Pripojeno!");
  Serial.println("IP adresa: "); Serial.println(WiFi.localIP());
  delay(100);
}

void WiFiOff() {
  Serial.println("Odpojuji server MQTT..., vypinam WiFi");
  delay(200);
  WiFi.disconnect();
  WiFi.mode(WIFI_OFF); 
  WiFi.forceSleepBegin();
  delay(1);
}

// Function to connect and reconnect as necessary to the MQTT server.
// Should be called in the loop function and it will take care if connecting.
void MQTT_connect() {
  int8_t ret;

  // Stop if already connected.
  if (mqtt.connected()) {
    return;
  }

  Serial.println("Pripojuji server MQTT... ");

  uint8_t retries = 3;
  while ((ret = mqtt.connect()) != 0) { // connect will return 0 for connected
       Serial.println(mqtt.connectErrorString(ret));
       Serial.println("Zkousim znovu navazat spojeni se serverem MQTT za 5 sekund ...");
       mqtt.disconnect();
       delay(5000);  // wait 5 seconds
       retries--;
       if (retries == 0) {
         // basically die and wait for WDT to reset me
         while (1);
       }
  }
  Serial.println("MQTT Pripojeno!");
}

//Calculates each of the variables that wunderground is expecting
void vypoctiHodnoty()
{
    //Calc winddir
    winddir = get_wind_direction();

    //Calc windspdms_avg5m
    float temp = 0;
    for(int i = 0 ; i < WIND_DIR_AVG_SIZE ; i++) {
        temp += windspdavg[i];
    }
      
    temp /= WIND_DIR_AVG_SIZE;
    windspdms_avg5m = temp;

    //Calc winddir_avg5m, Wind Direction
    //You can't just take the average. Google "mean of circular quantities" for more info
    //We will use the Mitsuta method because it doesn't require trig functions
    //And because it sounds cool.
    //Based on: http://abelian.org/vlf/bearings.html
    //Based on: http://stackoverflow.com/questions/1813483/averaging-angles-again
    long sum = winddiravg[0];
    int D = winddiravg[0];
    for(int i = 1 ; i < WIND_DIR_AVG_SIZE ; i++)
    {
        int delta = winddiravg[i] - D;

        if(delta < -180)
            D += delta + 360;
        else if(delta > 180)
            D += delta - 360;
        else
            D += delta;

        sum += D;
    }
    
    winddir_avg5m = sum / WIND_DIR_AVG_SIZE;
    if(winddir_avg5m >= 360) winddir_avg5m -= 360;
    if(winddir_avg5m < 0) winddir_avg5m += 360;

    //Calc windgustms_5m
    //Calc windgustdir_5m
    //Find the largest windgust in the last 5 minutes
    windgustms_5m = 0;
    windgustdir_5m = 0;

    //Step through the 5 minutes
    for(int i = 0; i < 5 ; i++)
    {
        if(windgust_5m[i] > windgustms_5m)
        {
            windgustms_5m = windgust_5m[i];
            windgustdir_5m = windgustdirection_5m[i];
        }
    }

    //Nactení hodnot z čidla teploty a vlhkosti
    humidity = dht.readHumidity();
    tempC = dht.readTemperature();  

    // Pokud se nezjistí vynulují se
    if (isnan(humidity)) {
      humidity = 0.0;
    }       
    if (isnan(tempC)) {
      tempC = 0.0;
    }       

}

//Calculate windspeed in m/s
float get_wind_speed()
{
    float deltaTime = millis() - lastWindCheck; //750ms

    deltaTime /= 1000.0; //Covert to seconds

    float windSpeed = (float)windClicks / deltaTime; //3 / 0.750s = 4

    windClicks = 0; //Reset and start watching for new wind
    lastWindCheck = millis();

    windSpeed *= 1.492; //4 * 1.492 = 5.968MPH
    windSpeed *= 0.44704; // 5.968 * 0.44704 = 2.66 m/s

    return(windSpeed);
}

//Calculate winddir
int get_wind_direction()
{
    unsigned int adc;

    adc = analogRead(WDIR); // get the current reading from the sensor

    float napeti= (float)(map(adc, 0, 1023, 0, 3300))/1000; // adcread je int, hodnota je tedy v mV
    //Serial.print("Vrácená hodnota na A0 " );Serial.println(napeti);

    // The following table is ADC readings for the wind direction sensor output, sorted from low to high.
    // Each threshold is the midpoint between adjacent headings. The output is degrees for that ADC reading.
    // Note that these are not in compass degree order! See Weather Meters datasheet for more information.
    
    if (napeti >= 3.04615384615385) return (270);
    if (napeti >= 2.85941255006676) return (315);
    if (napeti >= 2.6668465290028) return (292,5);
    if (napeti >= 2.53255813953488) return (0);
    if (napeti >= 2.26475149806133) return (337,5);
    if (napeti >= 2.03076923076923) return (225);
    if (napeti >= 1.93170731707317) return (247,5);
    if (napeti >= 1.48681318681319) return (45);
    if (napeti >= 1.30820392616466) return (22,5);
    if (napeti >= 0.925899280575539) return (180);
    if (napeti >= 0.787758224942617) return (202,5);
    if (napeti >= 0.595081967213115) return (135);
    if (napeti >= 0.406927277953435) return (157,5);
    if (napeti >= 0.3) return (90);
    if (napeti >= 0.270059880239521) return (67,5);
    if (napeti >= 0.212280701754386) return (112,5);

    return (-1); // error, disconnected?
}


//Send variables as mqtt message
void publikovatHodnotyPredSpanim()
{
    vypoctiHodnoty(); //Go calc all the various sensors

    WiFiOn();
    MQTT_connect();
    
    //ping the server to keep the mqtt connection alive
    if(!mqtt.ping()) {
      mqtt.disconnect();
    }
   
    //publikování hodnot na mqtt
    meteo_teplota.publish(tempC);
    meteo_vlhkost.publish(humidity); 
    meteo_vitr_smer.publish(winddir_avg5m);
    meteo_vitr_rychlost.publish(windspdms_avg5m);

    Serial.println("publikovatHodnotyPredSpanim");
 
    //Vytvoření JSON po odeslání informací o mě
    StaticJsonBuffer<200> jsonBuffer;
    JsonObject& vitrJSON = jsonBuffer.createObject();
  
    vitrJSON["windspdms_avg5m"] = windspdms_avg5m;
    vitrJSON["winddir_avg5m"] = winddir_avg5m;
    vitrJSON["windgustms_5m"] = windgustms_5m;
    vitrJSON["windgustdir_5m"] = windgustdir_5m;
    vitrJSON["humidity"] = humidity;
    vitrJSON["tempC"] = tempC;
  
    vitrJSON.printTo(MQTTMeteoVitrJSON);
    vitrJSON.printTo(Serial);
    Serial.println();
  
    meteo_vitr_json.publish(MQTTMeteoVitrJSON);

    WiFiOff();
}
